import logging
import re
from copy import deepcopy
from typing import Optional, Dict, List

import numpy as np
import pandas as pd


def detect_moisture_column(columns: List[str]) -> Optional[str]:
    """Detects the moisture column in a list of columns

    Args:
        columns: List of column names

    Returns:

    """
    res: Optional[str] = None
    search_regex: str = '(h2o)|(moisture)|(moist)|(mc)|(moisture_content)'
    for col in columns:
        if re.search(search_regex, col, re.IGNORECASE):
            res = col
            break
    return res


def solve_mass_moisture(mass_wet: pd.Series = None,
                        mass_dry: pd.Series = None,
                        moisture: pd.Series = None,
                        moisture_column_name: str = 'h2o',
                        rtol: float = 1e-05,
                        atol: float = 1e-08) -> pd.Series:
    logger = logging.getLogger(name=__name__)
    _vars: Dict = {k: v for k, v in deepcopy(locals()).items()}
    key_columns = ['mass_wet', 'mass_dry', 'moisture']
    vars_supplied: List[str] = [k for k in key_columns if _vars.get(k) is not None]

    if len(vars_supplied) == 3:
        logger.info('Over-specified - checking for balance.')
        re_calc_moisture = (mass_wet - mass_dry) / mass_wet * 100
        if not np.isclose(re_calc_moisture, moisture, rtol=rtol, atol=atol).all():
            msg = f"Mass balance is not satisfied: {re_calc_moisture}"
            logger.error(msg)
            raise ValueError(msg)
    elif len(vars_supplied) == 1:
        raise ValueError('Insufficient arguments supplied - at least 2 required.')

    var_to_solve: str = next((k for k, v in _vars.items() if v is None), None)

    res: Optional[pd.Series] = None
    if var_to_solve:
        calculations = {
            'mass_wet': lambda: mass_dry / (1 - moisture / 100),
            'mass_dry': lambda: mass_wet - (mass_wet * moisture / 100),
            'moisture': lambda: (mass_wet - mass_dry) / mass_wet * 100
        }

        res = calculations[var_to_solve]()
        res.name = var_to_solve if var_to_solve != 'moisture' else moisture_column_name  # use the supplied column name

    return res


_blockmodel_imports = None


# Define the Extras class to encapsulate the imported modules
class BlockmodelExtras:
    def __init__(self, omfpandas, omfvista, pv):
        self.omfpandas = omfpandas
        self.omfvista = omfvista
        self.pv = pv


def import_blockmodel_packages():
    """Helper method to safely import (only once) the blockmodel packages."""
    global _blockmodel_imports

    # Optional imports
    try:
        import omfpandas
        import omfvista
        import pyvista as pv
        from pyvista import CellType
    except ImportError as e:
        raise ImportError("Optional packages omfpandas or omfvista is not installed."
                          "Please install it to use this feature.") from e

    if _blockmodel_imports is None:
        try:
            import omfpandas
            import omfvista
            import pyvista as pv
            _blockmodel_imports = (omfpandas, omfvista, pv)
        except ImportError:
            raise ImportError("Failed to import blockmodel related packages. "
                              "Consider executing: 'poetry install --extras blockmodel'")
    return _blockmodel_imports



import logging

import pandas as pd

from elphick.geomet.base import MassComposition
from elphick.geomet.flowsheet import Flowsheet
from elphick.geomet.flowsheet.stream import Stream
from elphick.geomet.utils.moisture import solve_mass_moisture
from elphick.geomet.utils.pandas import composition_to_mass


def coerce_estimates(estimate_stream: Stream, input_stream: Stream,
                     recovery_bounds: tuple[float, float] = (0.01, 0.99),
                     complement_name: str = 'complement',
                     fs_name: str = 'Flowsheet',
                     show_plot: bool = False) -> Stream:
    """Coerce output estimates within recovery and the component range.

    Estimates contain error and at times can exceed the specified component range, or can consume more component
    mass than is available in the feed.  This function modifies (coerces) only the non-compliant estimate records
    in order to balance the node and keep all dry components within range.  Moisture is not modified.

            estimate_stream (supplied)
                /
    input_stream (supplied)
                \
            complement_stream

    1. limits the estimate to within the recovery bounds,
    2. ensures the estimate is within the component range,
    3. solves the complement, and ensures it is in range,
    4. if the complement is out of range, it is adjusted and the estimate adjusted to maintain the balance.

    Args:
        estimate_stream: The estimated object, which is a node output
        input_stream: The input object, which is a node input
        recovery_bounds: The bounds for the recovery, default is 0.01 to 0.99
        complement_name: The name of the complement , for plots.
        fs_name: The name of the flowsheet, for plots.
        show_plot: If True, show the network plot

    Returns:
        The coerced estimate stream
    """

    if show_plot:
        complement_stream: MassComposition = input_stream.sub(estimate_stream, name=complement_name)
        fs: Flowsheet = Flowsheet.from_objects([input_stream, estimate_stream, complement_stream],
                                               name=f"{fs_name}: Balance prior to coercion")
        fs.table_plot(plot_type='network', table_area=0.2, table_pos='top').show()

        # # debugging snippet to show a failing record
        # qry: str = 'index==1000'
        # fs_debug: Flowsheet = fs.query(qry)
        # fs_debug.name = f"{fs_name}: Balance prior to coercion: [{qry}]"
        # fs_debug.table_plot(plot_type='network', table_area=0.2, table_pos='top').show()

    if input_stream.status.ok is False:
        raise ValueError('Input stream is not OK')

    # clip the composition to the bounds
    estimate_stream = estimate_stream.clip_composition()

    # coerce the estimate component mass to within the total dry mass
    estimate_stream = estimate_stream.balance_composition()

    # clip the recovery
    estimate_stream = estimate_stream.clip_recovery(other=input_stream, recovery_bounds=(0.01, 0.99))

    if estimate_stream.status.ok is False:
        raise ValueError('Estimate stream is not OK - it should be after bounding recovery')

    # solve the complement
    complement_stream: Stream = input_stream.sub(estimate_stream, name=complement_name)

    # clip the composition to the bounds
    complement_stream = complement_stream.clip_composition()

    # coerce the estimate component mass to within the total dry mass
    complement_stream = complement_stream.balance_composition()

    # adjust the estimate to maintain the balance
    estimate_stream = input_stream.sub(complement_stream, name=estimate_stream.name,
                                       include_supplementary_data=True)

    if estimate_stream.status.ok is False:
        # This can occur in cases where the complement grade has been reduced (by balance_composition) to a point
        # where the resultant estimate grade is out of range.  In this case, we need to adjust the complement grade.

        estimate_stream = estimate_stream.clip_composition()
        complement_stream = input_stream.sub(estimate_stream, name=complement_name)

        if estimate_stream.status.ok is False:
            raise ValueError('Estimate stream is not OK after adjustment')

    fs2: Flowsheet = Flowsheet.from_objects([input_stream, estimate_stream, complement_stream],
                                            name=f"{fs_name}: Coerced Estimates")

    if show_plot:
        fs2.table_plot(plot_type='network', table_area=0.2, table_pos='top').show()

    if fs2.all_nodes_healthy is False:
        if fs2.all_streams_healthy and not fs2.all_nodes_healthy:
            logging.warning('All streams are healthy but not all nodes are healthy.  Consider the water balance.')
        else:
            raise ValueError('Flowsheet is not balanced after adjustment')

    return estimate_stream



from joblib import Parallel
from tqdm import tqdm


class TqdmParallel(Parallel):
    def __init__(self, *args, **kwargs):
        self._desc = kwargs.pop('desc', None)  # Get the description from kwargs
        self._tqdm = tqdm(total=kwargs.pop('total', None), desc=self._desc)  # Pass the description to tqdm
        super().__init__(*args, **kwargs)

    def __call__(self, iterable):
        iterable = list(iterable)
        self._tqdm.total = len(iterable)
        result = super().__call__(iterable)
        self._tqdm.close()
        return result

    def _print(self, msg, *msg_args):
        return

    def print_progress(self):
        self._tqdm.update()

    def _dispatch(self, batch):
        job_idx = super()._dispatch(batch)
        return job_idx

    def _collect(self, output):
        return super()._collect(output)



import logging
from functools import wraps
from pathlib import Path
from typing import Optional, Union, Literal, TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy import stats

from elphick.geomet import extras
from elphick.geomet.base import MassComposition
from elphick.geomet.extras import BlockmodelExtras
from elphick.geomet.utils.timer import log_timer

if TYPE_CHECKING:
    import pyvista as pv


# Modify the import_extras decorator
def import_extras(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        omfpandas, omfvista, pv = extras.import_blockmodel_packages()
        extras_instance = BlockmodelExtras(omfpandas, omfvista, pv)
        return func(*args, imports=extras_instance, **kwargs)

    return wrapper


class BlockModel(MassComposition):
    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 name: Optional[str] = None,
                 moisture_in_scope: bool = True,
                 mass_wet_var: Optional[str] = None,
                 mass_dry_var: Optional[str] = None,
                 moisture_var: Optional[str] = None,
                 component_vars: Optional[list[str]] = None,
                 composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                 components_as_symbols: bool = True,
                 ranges: Optional[dict[str, list]] = None,
                 config_file: Optional[Path] = None):

        if data is not None:
            if isinstance(data.index, pd.MultiIndex):
                if all([n.lower() in data.index.names for n in ['x', 'y', 'z', 'dx', 'dy', 'dz']]):
                    self.is_irregular = True
                elif all([n.lower() in data.index.names for n in ['x', 'y', 'z']]):
                    self.is_irregular = False
                data.index.set_names([n.lower() for n in data.index.names], inplace=True)

            else:
                raise ValueError("The index must be a pd.MultiIndex with names ['x', 'y', 'z'] "
                                 "or [['x', 'y', 'z', 'dx', 'dy', 'dz'].")

            # sort the data to ensure consistent with pyvista
            data.sort_index(level=['z', 'y', 'x'], ascending=[True, True, True], inplace=True)

        super().__init__(data=data, name=name, moisture_in_scope=moisture_in_scope,
                         mass_wet_var=mass_wet_var, mass_dry_var=mass_dry_var,
                         moisture_var=moisture_var, component_vars=component_vars,
                         composition_units=composition_units, components_as_symbols=components_as_symbols,
                         ranges=ranges, config_file=config_file)

    @classmethod
    @import_extras
    def from_omf(cls, omf_filepath: Path, imports,
                 element_name: Optional[str] = None,
                 columns: Optional[list[str]] = None,
                 query: Optional[str] = None,
                 density: float = 2.5) -> 'BlockModel':
        """Create a BlockModel from an OMF file.

        Args:
            omf_filepath: Path to the OMF file.
            imports: internally used to import the necessary packages.
            element_name: The name of the element in the OMF file.
            columns: The columns to extract from the OMF file.
            query: The query to filter the DataFrame.
            density: The density of the material in g/cm3, used to calculate DMT (Dry Mass Tonnes).  A workaround.

        Returns:
            BlockModel: The BlockModel instance.

        """

        omfpr: imports.omfpandas.OMFPandasReader = imports.omfpandas.OMFPandasReader(filepath=omf_filepath)
        blocks: pd.DataFrame = omfpr.read_blockmodel(blockmodel_name=element_name, attributes=columns,
                                                     query=query)

        # get the block volume

        volume: Union[float, np.ndarray[float]]
        from omfpandas.blockmodel import OMFBlockModel
        from omfpandas.blockmodels.convert_blockmodel import df_to_blockmodel
        from omfpandas.blockmodels.geometry import Geometry
        geom: Geometry = OMFBlockModel(df_to_blockmodel(blocks, blockmodel_name=element_name)).geometry

        if geom.__class__.__name__ == 'RegularGeometry':
            volume = geom.block_size[0] * geom.block_size[1] * geom.block_size[2]
        elif geom.__class__.__name__ == 'TensorGeometry':
            # TODO: Implement the volume calculation for TensorGeometry - this is a placeholder.
            volume = geom.block_sizes[0][0] * geom.block_sizes[0][1] * geom.block_sizes[0][2]
        else:
            raise ValueError(f"Geometry type '{geom.__class__.__name__}' not supported.")

        if density is not None:
            blocks['mass_dry'] = volume * density
            moisture_in_scope = False

        return cls(data=blocks, name=element_name, mass_dry_var='mass_dry', moisture_in_scope=moisture_in_scope)

    @import_extras
    def to_omf(self, omf_filepath: Path, imports, element_name: str = 'Block Model',
               description: str = 'A block model'):
        """Write the BlockModel to an OMF file.

        Args:
            omf_filepath: Path to the OMF file.
            imports: internally used to import the necessary packages.
            element_name: The name of the element in the OMF file.
            description: Description of the block model.
        """
        # Create an OMFPandasWriter instance
        writer = imports.omfpandas.OMFPandasWriter(filepath=omf_filepath)

        # Write the block model to the OMF file
        writer.write_blockmodel(blockmodel_name=element_name, description=description, dataframe=self.data)

    @log_timer
    def get_blocks(self) -> Union['pv.StructuredGrid', 'pv.UnstructuredGrid']:

        try:
            # Attempt to create a regular grid
            grid = self.create_structured_grid()
            self._logger.debug("Created a pv.StructuredGrid.")
        except ValueError:
            # If it fails, create an irregular grid
            grid = self.create_unstructured_grid()
            self._logger.debug("Created a pv.UnstructuredGrid.")
        return grid

    @import_extras
    def plot(self, scalar: str, imports, show_edges: bool = True) -> 'pv.Plotter':

        if scalar not in self.data_columns:
            raise ValueError(f"Column '{scalar}' not found in the DataFrame.")

        # Create a PyVista plotter
        plotter = imports.pv.Plotter()

        mesh = self.get_blocks()

        # Add a thresholded mesh to the plotter
        plotter.add_mesh_threshold(mesh, scalars=scalar, show_edges=show_edges)

        return plotter

    def is_regular(self) -> bool:
        """
        Determine if the grid spacing is complete and regular
        If it is, a pv.StructuredGrid is suitable.
        If not, a pv.UnstructuredGrid is suitable.

        :return:
        """

        block_sizes = np.array(self._block_sizes())
        return np.all(np.isclose(np.mean(block_sizes, axis=1), 0))

    def _block_sizes(self):
        data = self.data
        x_unique = data.index.get_level_values('x').unique()
        y_unique = data.index.get_level_values('y').unique()
        z_unique = data.index.get_level_values('z').unique()

        x_spacing = np.diff(x_unique)
        y_spacing = np.diff(y_unique)
        z_spacing = np.diff(z_unique)

        return x_spacing, y_spacing, z_spacing

    def common_block_size(self):
        data = self.data
        x_unique = data.index.get_level_values('x').unique()
        y_unique = data.index.get_level_values('y').unique()
        z_unique = data.index.get_level_values('z').unique()

        x_spacing = np.abs(np.diff(x_unique))
        y_spacing = np.abs(np.diff(y_unique))
        z_spacing = np.abs(np.diff(z_unique))

        return stats.mode(x_spacing).mode, stats.mode(y_spacing).mode, stats.mode(z_spacing).mode

    @import_extras
    def create_structured_grid(self, imports) -> 'pv.StructuredGrid':

        # Get the unique x, y, z coordinates (centroids)
        data = self.data.sort_values(['z', 'y', 'x'])  # ensure the data is sorted F-style
        x_centroids = data.index.get_level_values('x').unique()
        y_centroids = data.index.get_level_values('y').unique()
        z_centroids = data.index.get_level_values('z').unique()

        # Calculate the cell size (assuming all cells are of equal size)
        dx = np.diff(x_centroids)[0]
        dy = np.diff(y_centroids)[0]
        dz = np.diff(z_centroids)[0]

        # Calculate the grid points
        x_points = np.concatenate([x_centroids - dx / 2, x_centroids[-1:] + dx / 2])
        y_points = np.concatenate([y_centroids - dy / 2, y_centroids[-1:] + dy / 2])
        z_points = np.concatenate([z_centroids - dz / 2, z_centroids[-1:] + dz / 2])

        # Create the 3D grid of points
        x, y, z = np.meshgrid(x_points, y_points, z_points, indexing='ij')

        # Create a StructuredGrid object
        grid = imports.pv.StructuredGrid(x, y, z)

        # Add the data from the DataFrame to the grid
        for column in data.columns:
            grid.cell_data[column] = data[column].values

        return grid

    def create_voxels(self) -> 'pv.UnstructuredGrid':
        grid = self.voxelise(self.data)
        return grid

    @import_extras
    def create_unstructured_grid(self, imports) -> 'pv.UnstructuredGrid':
        """
        Requires the index to be a pd.MultiIndex with names ['x', 'y', 'z', 'dx', 'dy', 'dz'].
        :return:
        """

        # Get the x, y, z coordinates and cell dimensions
        blocks = self.data.reset_index().sort_values(['z', 'y', 'x'])  # ensure the data is sorted F-style
        # if no dims are passed, estimate them
        if 'dx' not in blocks.columns:
            dx, dy, dz = self.common_block_size()
            blocks['dx'] = dx
            blocks['dy'] = dy
            blocks['dz'] = dz

        x, y, z, dx, dy, dz = (blocks[col].values for col in blocks.columns if col in ['x', 'y', 'z', 'dx', 'dy', 'dz'])
        blocks.set_index(['x', 'y', 'z', 'dx', 'dy', 'dz'], inplace=True)
        # Create the cell points/vertices
        # REF: https://github.com/OpenGeoVis/PVGeo/blob/main/PVGeo/filters/voxelize.py

        n_cells = len(x)

        # Generate cell nodes for all points in data set
        # - Bottom
        c_n1 = np.stack(((x - dx / 2), (y - dy / 2), (z - dz / 2)), axis=1)
        c_n2 = np.stack(((x + dx / 2), (y - dy / 2), (z - dz / 2)), axis=1)
        c_n3 = np.stack(((x - dx / 2), (y + dy / 2), (z - dz / 2)), axis=1)
        c_n4 = np.stack(((x + dx / 2), (y + dy / 2), (z - dz / 2)), axis=1)
        # - Top
        c_n5 = np.stack(((x - dx / 2), (y - dy / 2), (z + dz / 2)), axis=1)
        c_n6 = np.stack(((x + dx / 2), (y - dy / 2), (z + dz / 2)), axis=1)
        c_n7 = np.stack(((x - dx / 2), (y + dy / 2), (z + dz / 2)), axis=1)
        c_n8 = np.stack(((x + dx / 2), (y + dy / 2), (z + dz / 2)), axis=1)

        # - Concatenate
        # nodes = np.concatenate((c_n1, c_n2, c_n3, c_n4, c_n5, c_n6, c_n7, c_n8), axis=0)
        nodes = np.hstack((c_n1, c_n2, c_n3, c_n4, c_n5, c_n6, c_n7, c_n8)).ravel().reshape(n_cells * 8, 3)

        # create the cells
        # REF: https://docs/pyvista.org/examples/00-load/create-unstructured-surface.html
        cells_hex = np.arange(n_cells * 8).reshape(n_cells, 8)

        grid = imports.pv.UnstructuredGrid({imports.pv.CellType.VOXEL: cells_hex}, nodes)

        # add the attributes (column) data
        for col in blocks.columns:
            grid.cell_data[col] = blocks[col].values

        return grid

    @staticmethod
    @log_timer
    @import_extras
    def voxelise(blocks, imports) -> 'pv.UnstructuredGrid':

        logger = logging.getLogger(__name__)
        msg = "Voxelising blocks requires PVGeo package."
        logger.error(msg)
        raise NotImplementedError(msg)

        # vtkpoints = PVGeo.points_to_poly_data(centroid_data)

        x_values = blocks.index.get_level_values('x').values
        y_values = blocks.index.get_level_values('y').values
        z_values = blocks.index.get_level_values('z').values

        # Stack x, y, z values into a numpy array
        centroids = np.column_stack((x_values, y_values, z_values))

        # Create a PolyData object
        polydata = imports.pv.PolyData(centroids)

        # Add cell values as point data
        for column in blocks.columns:
            polydata[column] = blocks[[column]]

        # Create a Voxelizer filter
        voxelizer = PVGeo.filters.VoxelizePoints()
        # Apply the filter to the points
        grid = voxelizer.apply(polydata)

        logger.info(f"Voxelised {blocks.shape[0]} points.")
        logger.info("Recovered Angle (deg.): %.3f" % voxelizer.get_angle())
        logger.info("Recovered Cell Sizes: (%.2f, %.2f, %.2f)" % voxelizer.get_spacing())

        return grid

    def __str__(self):
        return f"BlockModel: {self.name}\n{self.aggregate.to_dict()}"



"""
Pandas utils
"""
import inspect
import logging
import tokenize
from io import StringIO
from token import STRING
from typing import List, Dict, Optional, Literal

import numpy as np
import pandas as pd

from elphick.geomet.utils.components import is_compositional, get_components
from elphick.geomet.utils.moisture import solve_mass_moisture, detect_moisture_column
from elphick.geomet.utils.size import mean_size

composition_factors: dict[str, int] = {'%': 100, 'ppm': 1e6, 'ppb': 1e9}


def column_prefixes(columns: List[str]) -> Dict[str, List[str]]:
    return {prefix: [col for col in columns if prefix == col.split('_')[0]] for prefix in
            list(dict.fromkeys([col.split('_')[0] for col in columns if len(col.split('_')) > 1]))}


def column_prefix_counts(columns: List[str]) -> Dict[str, int]:
    return {k: len(v) for k, v in column_prefixes(columns).items()}


def mass_to_composition(df: pd.DataFrame,
                        mass_wet: Optional[str] = 'mass_wet',
                        mass_dry: str = 'mass_dry',
                        moisture_column_name: Optional[str] = None,
                        component_columns: Optional[list[str]] = None,
                        composition_units: Literal['%', 'ppm', 'ppb'] = '%') -> pd.DataFrame:
    """Convert a mass DataFrame to composition

    Supplementary columns (columns that are not mass or composition) are ignored.

    Args:
        df: The pd.DataFrame containing mass.  H2O if provided will be ignored.  All columns other than the
         mass_wet and mass_dry are assumed to be `additive`, that is, dry mass weighting is valid.
         Assumes composition is in %w/w units.
        mass_wet: The wet mass column, optional. If not provided, it's assumed to be equal to mass_dry.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        moisture_column_name: if mass_wet is provided, the resultant moisture will be returned with this column name.
         If None, and moisture is detected in the input, that column name will be used instead.

        component_columns: The composition columns to be used for the calculation.  If not provided, the columns
         will be auto-detected using a case in-sensitive match to all elements and oxides.  H2O is excluded
        composition_units: determines the factor to convert mass to composition.

    Returns:
        A pd.Dataframe containing mass (wet and dry mass) and composition
    """

    moisture_column_name, mass_moisture_cols, component_cols = prepare_columns(df, mass_wet, mass_dry,
                                                                               moisture_column_name, component_columns)

    if mass_wet and mass_wet in df.columns:
        mass: pd.DataFrame = df[[mass_wet, mass_dry]]
    else:
        mass: pd.DataFrame = df[[mass_dry]]

    component_mass: pd.DataFrame = df[component_cols]
    composition: pd.DataFrame = component_mass.div(mass[mass_dry].replace(0.0, np.nan), axis=0).fillna(0.0) * composition_factors[composition_units]

    if mass_wet and (mass_wet in df.columns):
        moisture: pd.Series = solve_mass_moisture(mass_wet=mass[mass_wet], mass_dry=mass[mass_dry]).rename(
            moisture_column_name)
        return pd.concat([mass, moisture, composition], axis='columns')
    else:
        return pd.concat([mass, composition], axis=1)


def composition_to_mass(df: pd.DataFrame,
                        mass_wet: Optional[str] = None,
                        mass_dry: str = 'mass_dry',
                        component_columns: Optional[list[str]] = None,
                        moisture_column_name: Optional[str] = None,
                        composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                        return_moisture: bool = False) -> pd.DataFrame:
    """ Convert a composition DataFrame to mass

        Supplementary columns (columns that are not mass or composition) are ignored.

    Args:
        df: The pd.DataFrame containing mass.  H2O if provided will be ignored.  All columns other than the
         mass_wet and mass_dry are assumed to be `additive`, that is, dry mass weighting is valid.
         Assumes composition is in %w/w units.
        mass_wet: The wet mass column, optional.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        moisture_column_name: if mass_wet is provided, the resultant moisture will be returned with this column name.
         If None, and moisture is detected in the input, that column name will be used instead.
        component_columns: The composition columns to be used for the calculation.  If not provided, the columns
         will be auto-detected using a case in-sensitive match to all elements and oxides.  H2O is excluded
        composition_units: determines the factor to convert composition to mass.
        return_moisture: If True, the moisture column will be returned.

    Returns:
        A pd.Dataframe containing the mass representation of mass totals and components
    """

    moisture_column_name, mass_moisture_cols, component_cols = prepare_columns(df, mass_wet, mass_dry,
                                                                               moisture_column_name, component_columns)

    if mass_wet and mass_wet in df.columns:
        mass: pd.DataFrame = df[[mass_wet, mass_dry]]
    else:
        mass: pd.DataFrame = df[[mass_dry]]

    composition: pd.DataFrame = df[component_cols]
    component_mass: pd.DataFrame = composition.mul(mass[mass_dry], axis=0) / composition_factors[composition_units]

    if mass_wet and (mass_wet in df.columns) and return_moisture:
        moisture: pd.Series = (mass[mass_wet] - mass[mass_dry]).rename(moisture_column_name)
        return pd.concat([mass, moisture, component_mass], axis='columns')
    else:
        return pd.concat([mass, component_mass], axis=1)


def prepare_columns(df: pd.DataFrame, mass_wet: Optional[str], mass_dry: str, moisture_column_name: Optional[str],
                    component_columns: Optional[list[str]]) -> tuple[str, List[str], List[str]]:
    if moisture_column_name is None:
        moisture_column_name = detect_moisture_column(df.columns)
        # if moisture_column_name is None:
        #     moisture_column_name = 'h2o'  # set default value to 'h2o' if not detected
    mass_moisture_cols = [mass_wet, mass_dry, moisture_column_name]

    if component_columns is None:
        non_mass_cols: list[str] = [col for col in df.columns if col.lower() not in mass_moisture_cols]
        component_cols: list[str] = get_components(df[non_mass_cols], strict=False)
    else:
        component_cols: list[str] = component_columns

    return moisture_column_name, mass_moisture_cols, component_cols


def weight_average(df: pd.DataFrame,
                   mass_wet: Optional[str] = None,
                   mass_dry: str = 'mass_dry',
                   moisture_column_name: Optional[str] = None,
                   component_columns: Optional[list[str]] = None,
                   composition_units: Literal['%', 'ppm', 'ppb'] = '%') -> pd.Series:
    """Weight Average a DataFrame containing mass-composition

    Args:
        df: The pd.DataFrame containing mass-composition.  H2O if provided will be ignored.  All columns other than the
         mass_wet and mass_dry are assumed to be `additive`, that is, dry mass weighting is valid.
         Assumes composition is in %w/w units.
        mass_wet: The optional wet mass column.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        moisture_column_name: if mass_wet is provided, the resultant moisture will be returned with this column name.
         If None, and moisture is detected in the input, that column name will be used instead.
        component_columns: The composition columns to be used for the calculation.  If not provided, the columns
         will be auto-detected using a case in-sensitive match to all elements and oxides.  H2O is excluded
        composition_units: determines the factor to convert mass to composition.

    Returns:
        A pd.Series containing the total mass and weight averaged composition.
    """
    moisture_column_name, mass_moisture_cols, component_cols = prepare_columns(df, mass_wet, mass_dry,
                                                                               moisture_column_name, component_columns)

    mass_sum: pd.DataFrame = df.pipe(composition_to_mass, mass_wet=mass_wet, mass_dry=mass_dry,
                                     moisture_column_name=moisture_column_name,
                                     component_columns=component_columns,
                                     composition_units=composition_units).sum(axis="index").to_frame().T

    component_cols = [col for col in component_cols if
                      col.lower() not in [mass_wet, mass_dry, 'h2o', 'moisture']]

    weighted_composition: pd.Series = mass_sum[component_cols].div(mass_sum[mass_dry], axis=0) * composition_factors[
        composition_units]

    if mass_wet and (mass_wet in df.columns):
        moisture: pd.Series = solve_mass_moisture(mass_wet=mass_sum[mass_wet], mass_dry=mass_sum[mass_dry])
        return pd.concat([mass_sum[[mass_wet, mass_dry]], moisture, weighted_composition], axis=1).iloc[0].rename(
            'weight_average')
    else:
        return pd.concat([mass_sum[[mass_dry]], weighted_composition], axis=1).iloc[0].rename('weight_average')


def calculate_recovery(df: pd.DataFrame,
                       df_ref: pd.DataFrame,
                       mass_wet: str = 'mass_wet',
                       mass_dry: str = 'mass_dry') -> pd.DataFrame:
    """Calculate recovery of mass-composition for two DataFrames

    Args:
        df: The pd.DataFrame containing mass-composition.  H2O if provided will be ignored.  All columns other than the
         mass_wet and mass_dry are assumed to be `additive`, that is, dry mass weighting is valid.
         Assumes composition is in %w/w units.
        df_ref: The stream that df will be divided by to calculate the recovery.  Often the feed stream.
        mass_wet: The wet mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.

    Returns:
        A pd.Series containing the total mass and weight averaged composition.
    """

    res: pd.DataFrame = df.pipe(composition_to_mass, mass_wet=mass_wet, mass_dry=mass_dry) / df_ref.pipe(
        composition_to_mass, mass_wet=mass_wet, mass_dry=mass_dry)
    return res


def calculate_partition(df_feed: pd.DataFrame,
                        df_preferred: pd.DataFrame,
                        col_mass_dry: str = 'mass_dry') -> pd.DataFrame:
    """Calculate the partition curve from two streams

    .. math::
        K = \\frac{{m_{preferred}}}{{m_{feed}}}

    Applicable to the one dimensional case only.  The PN is bounded [0, 1].
    The interval mean for size is the geometric mean, otherwise the arithmetic mean.
    The interval mean is named `da`, which can be interpreted as `diameter-average` or `density-average`.
    TODO: consider a generalised name, fraction-average -> fa?

    Args:
        df_feed: The pd.DataFrame containing mass-composition representing the fractionated feed.
        df_preferred: The pd.DataFrame containing mass-composition representing the fractionated preferred stream.
        col_mass_dry: The dry mass column, not optional.

    Returns:
        A pd.DataFrame containing the partition data with a range [0, 1].
    """

    res: pd.DataFrame = df_preferred[[col_mass_dry]].div(df_feed[[col_mass_dry]]).rename(columns={col_mass_dry: 'K'})
    if df_preferred.index.name.lower() == 'size':
        res.insert(loc=0, column='size', value=mean_size(res.index))
    else:
        res.insert(loc=0, column=df_preferred.index.name.lower(), value=res.index.mid)
    return res


def cumulate(mass_data: pd.DataFrame, direction: str) -> pd.DataFrame:
    """Cumulate along the index

    Expected use case is only for Datasets that have been reduced to 1D.

    Args:
        mass_data: The mass data to cumulate - note composition must be represented as mass
        direction: 'ascending'|'descending'

    Returns:

    """

    valid_dirs: List[str] = ['ascending', 'descending']
    if direction not in valid_dirs:
        raise KeyError(f'Invalid direction provided.  Valid arguments are: {valid_dirs}')

    d_dir: Dict = {'ascending': True if direction == 'ascending' else False,
                   'descending': True if direction == 'descending' else False}

    if mass_data.index.ndim > 1:
        raise NotImplementedError('DataFrames having indexes > 1D have not been tested.')

    index_var: str = mass_data.index.name
    if not isinstance(mass_data.index, pd.IntervalIndex):
        raise NotImplementedError(f"The {index_var} of this object is not a pd.Interval. "
                                  f" Only 1D interval objects are valid")

    interval_index = mass_data.index.get_level_values(index_var)
    if not (interval_index.is_monotonic_increasing or interval_index.is_monotonic_decreasing):
        raise ValueError('Index is not monotonically increasing or decreasing')

    in_data_ascending: bool = True
    if interval_index.is_monotonic_decreasing:
        in_data_ascending = False

    # sort by the direction provided, first save the index
    original_index: pd.Index = mass_data.index
    try:
        mass_data: pd.DataFrame = mass_data.sort_index(ascending=d_dir[direction])
        mass_cum: pd.DataFrame = mass_data.cumsum()

    finally:
        # reset the index to the original
        mass_data = mass_data.reindex(original_index)

    return mass_cum


def _detect_non_float_columns(df):
    _logger: logging.Logger = logging.getLogger(inspect.stack()[1].function)
    non_float_cols: List = [col for col in df.columns if col not in df.select_dtypes(include=[float, int]).columns]
    if len(non_float_cols) > 0:
        _logger.info(f"The following columns are not float columns and will be ignored: {non_float_cols}")
    return non_float_cols


def _detect_non_component_columns(df):
    _logger: logging.Logger = logging.getLogger(inspect.stack()[1].function)
    chemistry_vars = [col.lower() for col in is_compositional(df.columns, strict=False).values() if col not in ['H2O']]

    non_float_cols: List = [col for col in df.columns if
                            col not in (list(df.select_dtypes(include=[float, int]).columns) + chemistry_vars + [
                                'mass_wet', 'mass_dry', 'h2o'])]
    if len(non_float_cols) > 0:
        _logger.info(f"The following columns are not float columns and will be ignored: {non_float_cols}")
    return non_float_cols


class MeanIntervalIndex(pd.IntervalIndex):
    """MeanIntervalIndex is a subclass of pd.IntervalIndex that calculates the mean of the interval bounds."""

    def __new__(cls, data, mean_values=None):
        obj = pd.IntervalIndex.__new__(cls, data)
        return obj

    def __init__(self, data, mean_values=None):
        self.mean_values = mean_values

    @property
    def mean(self):
        if self.mean_values is not None:
            return self.mean_values
        elif self.name == 'size':
            # Calculate geometric mean
            return mean_size(self)
        else:
            # Calculate arithmetic mean
            return (self.right + self.left) / 2


# class MeanIntervalArray(pd.arrays.IntervalArray):
#     def __init__(self, data, dtype=None, copy=False):
#         super().__init__(data, dtype, copy)
#         if self.name == 'size':
#             # Calculate geometric mean
#             self.mean_values = gmean([self.right, self.left], axis=0)
#         else:
#             # Calculate arithmetic mean
#             self.mean_values = (self.right + self.left) / 2
#
#     @property
#     def mean(self):
#         if self.mean_values is not None:
#             return self.mean_values
#         elif self.name == 'size':
#             # Calculate geometric mean
#             return gmean([self.right, self.left], axis=0)
#         else:
#             # Calculate arithmetic mean
#             return (self.right + self.left) / 2


def parse_vars_from_expr(expr: str) -> list[str]:
    """ Parse variables from a pandas query expression string.

    Args:
        expr: The expression string

    Returns:
        list[str]: The list of variables
    """
    variables = set()
    tokens = tokenize.generate_tokens(StringIO(expr).readline)
    logical_operators = {'and', 'or', '&', '|'}
    inside_backticks = False
    current_var = []

    for token in tokens:
        if token.string == '`':
            if inside_backticks:
                # End of backtick-enclosed variable
                variables.add(' '.join(current_var))
                current_var = []
            inside_backticks = not inside_backticks
        elif inside_backticks:
            if token.type in {tokenize.NAME, STRING}:
                current_var.append(token.string)
        elif token.type == tokenize.NAME and token.string not in logical_operators:
            variables.add(token.string)

    return list(variables)



import logging
from typing import Dict, Optional, List, Union, Iterable, Tuple

import numpy as np
import pandas as pd
from joblib import delayed
from tqdm import tqdm

from elphick.geomet import Sample
from elphick.geomet.flowsheet.stream import Stream
# from elphick.geomet.utils.interp import _upsample_grid_by_factor
from elphick.geomet.utils.parallel import TqdmParallel
from elphick.geomet.utils.pandas import column_prefix_counts, column_prefixes

logger = logging.getLogger(__name__)


def create_stream(stream_data: Tuple[Union[int, str], pd.DataFrame],
                  interval_edges: Optional[Union[Iterable, int]] = None) -> list[Stream]:
    stream, data = stream_data
    res = None
    try:
        if interval_edges is not None:
            res = Stream(data=data, name=stream).resample_1d(interval_edges=interval_edges)
        else:
            res = Stream(data=data, name=stream)
    except Exception as e:
        logger.error(f"Error creating Sample object for {stream}: {e}")

    return res


def streams_from_dataframe(df: pd.DataFrame,
                           mc_name_col: Optional[str] = None,
                           interval_edges: Optional[Union[Iterable, int]] = None,
                           n_jobs=1) -> List[Sample]:
    """Objects from a DataFrame

    Args:
        df: The DataFrame
        mc_name_col: The column specified contains the names of objects to create.
          If None the DataFrame is assumed to be wide and the mc objects will be extracted from column prefixes.
        interval_edges: The values of the new grid (interval edges).  If an int, will up-sample by that factor, for
         example the value of 10 will automatically define edges that create 10 x the resolution (up-sampled).
         Applicable only to 1d interval indexes.
        n_jobs: The number of parallel jobs to run.  If -1, will use all available cores.

    Returns:
        List of Stream objects
    """
    stream_data: Dict[str, pd.DataFrame] = {}
    index_names: List[str] = []
    if mc_name_col:
        logger.debug("Creating Stream objects by name column.")
        if mc_name_col in df.index.names:
            index_names = df.index.names
            df.reset_index(mc_name_col, inplace=True)
        if mc_name_col not in df.columns:
            raise KeyError(f'{mc_name_col} is not in the columns or indexes.')
        names = df[mc_name_col].unique()
        for obj_name in tqdm(names, desc='Preparing Stream data'):
            stream_data[obj_name] = df.query(f'{mc_name_col} == @obj_name')[
                [col for col in df.columns if col != mc_name_col]]
        if index_names:  # reinstate the index on the original dataframe
            df.reset_index(inplace=True)
            df.set_index(index_names, inplace=True)
    else:
        logger.debug("Creating Stream objects by column prefixes.")
        # wide case - find prefixes where there are at least 3 columns
        prefix_counts = column_prefix_counts(df.columns)
        prefix_cols = column_prefixes(df.columns)
        for prefix, n in tqdm(prefix_counts.items(), desc='Preparing Stream data by column prefixes'):
            if n >= 3:  # we need at least 3 columns to create a Stream object
                logger.info(f"Creating object for {prefix}")
                cols = prefix_cols[prefix]
                stream_data[prefix] = df[[col for col in df.columns if col in cols]].rename(
                    columns={col: col.replace(f'{prefix}_', '') for col in df.columns})

    if interval_edges is not None:
        logger.debug("Resampling Stream objects to new interval edges.")
        # unify the edges - this will also interp missing grades
        if not isinstance(df.index, pd.IntervalIndex):
            raise NotImplementedError(f"The index `{df.index}` of the dataframe is not a pd.Interval. "
                                      f" Only 1D interval indexes are valid")
        if isinstance(interval_edges, int):
            raise NotImplementedError("Needs work on interp to convert from xr to pd")
            all_edges = []
            for strm_data in stream_data.values():
                all_edges.extend(list(np.sort(np.unique(list(strm_data.index.left) + list(strm_data.index.right)))))
            all_edges = list(set(all_edges))
            all_edges.sort()
            indx = pd.IntervalIndex.from_arrays(left=all_edges[0:-1], right=all_edges[1:])
            interval_edges = _upsample_grid_by_factor(indx=indx, factor=interval_edges)

    with TqdmParallel(desc="Creating Stream objects", n_jobs=n_jobs,
                      prefer=None, total=len(stream_data)) as p:
        res = p(delayed(create_stream)(stream_data, interval_edges) for stream_data in stream_data.items())

    return res



from elphick.geomet.datasets import Downloader
import pandas as pd


def load_a072391_assay(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='A072391_assay.zip', show_report=show_report)


def load_a072391_collars(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='A072391_collars.zip', show_report=show_report)


def load_a072391_geo(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='A072391_geo.zip', show_report=show_report)


def load_a072391_met(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='A072391_met.zip', show_report=show_report)


def load_a072391_wireline(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='A072391_wireline.zip', show_report=show_report)


def load_demo_data(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='demo_data.zip', show_report=show_report)


def load_iron_ore_sample_a072391(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='iron_ore_sample_A072391.zip', show_report=show_report)


def load_iron_ore_sample_xyz_a072391(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='iron_ore_sample_xyz_A072391.zip', show_report=show_report)


def load_nordic_iron_ore_sink_float(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='nordic_iron_ore_sink_float.zip', show_report=show_report)


def load_size_by_assay(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='size_by_assay.zip', show_report=show_report)


def load_size_distribution(show_report: bool = False) -> pd.DataFrame:
    return Downloader().load_data(datafile='size_distribution.zip', show_report=show_report)




from typing import List
import pandas as pd
from itertools import tee


def amenability_index(df_recovery: pd.DataFrame,
                      col_target: str,
                      col_mass_recovery: str):
    """Calculate the Amenability Index

    Implementation of the Amenability Index as presented in the paper titled APPLICATIONS OF INDIVIDUAL PARTICLE
    PYKNOMETRY by G. Elphick and Dr. T.F. Mason at the DMS POWDERS’ 10th FERROSILICON CONFERENCE.

    The amenability Index for a particular gangue analyte is the complement of the relative recovery to the target
     analyte across the full sample.  It is process / operating point independent, hence characterises the ore,
     not the process.

    Args:
        df_recovery: DataFrame containing the ideal incremental recovery of a fractionated sample.
        col_target: The column name of the target analyte
        col_mass_recovery: The column name of the mass_recovery (yield) column

    Returns:
        A pd.Series containing the Amenability Indices for the gangue analytes
    """
    cols: List[str] = [col for col in df_recovery.columns if col not in [col_mass_recovery]]
    area_target = area_trapezoid(xs=df_recovery[col_mass_recovery], ys=df_recovery[col_target])
    results: List = []
    for analyte in cols:
        area = area_trapezoid(xs=df_recovery[col_mass_recovery], ys=df_recovery[analyte])
        results.append(1 - (area / area_target))
    return pd.Series(results, index=cols, name='amenability_index')


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ...
    For use in py39, after which itertools.pairwise can be used
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def area_trapezoid(xs, ys):
    area = 0
    for (ax, ay), (bx, by) in pairwise(zip(xs, ys)):
        h = bx - ax
        area += h * (ay + by) / 2
    return area



from typing import Optional

import pandas as pd

import plotly.graph_objects as go


def plot_parallel(data: pd.DataFrame, color: Optional[str] = None, title: Optional[str] = None) -> go.Figure:
    """Create an interactive parallel plot

    Useful to explore multi-dimensional data like mass-composition data

    Args:
        data: Dataframe to plot
        color: Optional color variable
        title: Optional plot title

    Returns:

    """

    # Kudos: https://stackoverflow.com/questions/72125802/parallel-coordinate-plot-in-plotly-with-continuous-
    # and-categorical-data

    categorical_columns = data.select_dtypes(include=['category', 'object'])
    col_list = []

    for col in data.columns:
        if col in categorical_columns:  # categorical columns
            values = data[col].unique()
            value2dummy = dict(zip(values, range(
                len(values))))  # works if values are strings, otherwise we probably need to convert them
            data[col] = [value2dummy[v] for v in data[col]]
            col_dict = dict(
                label=col,
                tickvals=list(value2dummy.values()),
                ticktext=list(value2dummy.keys()),
                values=data[col],
            )
        else:  # continuous columns
            col_dict = dict(
                range=(data[col].min(), data[col].max()),
                label=col,
                values=data[col],
            )
        col_list.append(col_dict)

    if color is None:
        fig = go.Figure(data=go.Parcoords(dimensions=col_list))
    else:
        fig = go.Figure(data=go.Parcoords(dimensions=col_list, line=dict(color=data[color])))

    fig.update_layout(title=title)

    return fig




"""
Managing components/composition
"""

from typing import List, Dict, Union

import periodictable as pt
from periodictable.formulas import Formula

custom_components: List[str] = ['LOI', 'Ash']

# Kudos: pyrolite
DEFAULT_CHARGES: Dict = dict(
    H=1,
    Li=1,
    Be=1,
    B=3,
    C=4,
    O=-2,
    F=-1,
    Na=1,
    Mg=2,
    Al=3,
    Si=4,
    P=3,
    Cl=-1,
    K=1,
    Ca=2,
    Sc=3,
    Ti=4,
    V=3,
    Cr=3,
    Mn=2,
    Fe=2,
    Co=2,
    Ni=2,
    Cu=2,
    Zn=2,
    Br=-1,
    Rb=1,
    Sr=2,
    Y=3,
    Zr=4,
    Nb=5,
    Sn=4,
    I=-1,
    Cs=1,
    Ba=2,
    La=3,
    Ce=3,
    Pr=3,
    Nd=3,
    Sm=3,
    Eu=3,
    Gd=3,
    Tb=3,
    Dy=3,
    Ho=3,
    Er=3,
    Tm=3,
    Yb=3,
    Lu=3,
    Hf=4,
    Pb=2,
    Th=4,
    U=4,
)


def elements() -> List[str]:
    res: List[str] = [el.symbol for el in pt.elements]
    return res


def is_element(candidates: List[str], strict: bool = True) -> Union[List[str], Dict[str, str]]:
    if strict:
        matches: list = list(set(candidates).intersection(elements()))
    else:
        e_map: Dict[str, str] = {e.symbol.lower(): e.symbol for e in pt.elements}
        matches: Dict[str, str] = {c: e_map[c.lower()] for c in candidates if c.lower() in e_map.keys()}

    return matches


def oxides() -> List[Formula]:
    # cats = {e for e in [el for el in pt.elements if str(el) in DEFAULT_CHARGES.keys()] if DEFAULT_CHARGES[str(e)] > 0}
    cats = {el for el in pt.elements if (str(el) in DEFAULT_CHARGES.keys()) and (DEFAULT_CHARGES[str(el)] > 0)}

    res: List[Formula] = []
    for c in cats:
        charge = DEFAULT_CHARGES[str(c)]
        if charge % 2 == 0:
            res.append(pt.formula(str(c) + str(1) + 'O' + str(charge // 2)))
        else:
            res.append(pt.formula(str(c) + str(2) + 'O' + str(charge)))

    return res


def is_oxide(candidates: List[str], strict: bool = True) -> Union[List[str], Dict[str, str]]:
    if strict:
        oxs = {str(o) for o in oxides()}
        matches: list = list(set(candidates).intersection(oxs))
    else:
        o_map: Dict[str, str] = {str(o).lower(): str(o) for o in oxides()}
        matches: Dict[str, str] = {c: o_map[c.lower()] for c in candidates if c.lower() in o_map.keys()}

    return matches


def is_compositional(candidates: List[str], strict: bool = True) -> Union[List[str], Dict[str, str]]:
    """
    Check if a list of candidates are compositional components (elements or oxides)
    Args:
        candidates: list of string candidates
        strict: If True, the candidates must be in the list of known compositional components (elements or oxides)
        as chemical symbols.

    Returns:
        If strict, a list of compositional components, otherwise a dict of the original candidates (keys) and
        their compositional component symbols (values)
    """
    if strict:
        comps = {str(o) for o in oxides()}.union(set(elements())).union(set(custom_components))
        matches: list = list(set(candidates).intersection(comps))
    else:
        comp_map: Dict[str, str] = {**{str(o).lower(): str(o) for o in oxides()},
                                    **{a.lower(): a for a in elements()},
                                    **{c.lower(): c for c in custom_components}}
        matches: Dict[str, str] = {c: comp_map[c.lower()] for c in candidates if c.lower() in comp_map.keys()}

    return matches


def get_components(candidates: List[str], strict: bool = True) -> list[str]:
    return list(is_compositional(candidates, strict=strict).keys())
